# Copyright (c) 2011 John Stowers
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSES/GPL-3.0

import gi
gi.require_version("GnomeDesktop", "4.0")
from gi.repository import Gtk, GnomeDesktop, Gtk

from gtweak.gshellwrapper import GnomeShellFactory
from gtweak.widgets import TweakPreferencesPage, GSettingsTweakSwitchRow, GSettingsSwitchTweakValue, _GSettingsTweak, TweakPreferencesGroup, build_label_beside_widget, Tweak
from gtweak.tweakmodel import Tweak, TweakGroup
from gtweak.gsettings import GSettingsSetting, GSettingsMissingError



_shell = GnomeShellFactory().get_shell()
_shell_loaded = _shell is not None


class _XkbOption(Gtk.Expander, Tweak):
    def __init__(self, group_id, parent_settings, xkb_info, **options):
        try:
            desc = xkb_info.description_for_group(group_id)
        except AttributeError:
            desc = group_id
        Gtk.Expander.__init__(self)
        Tweak.__init__(self, desc, desc, **options)

        self.set_label(self.title)
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.set_margin_start(15)
        self.set_child(vbox)

        self._multiple_selection = group_id not in { 'keypad', 'kpdl', 'caps', 'altwin', 'nbsp', 'esperanto' }
        self._group_id = group_id
        self._parent_settings = parent_settings
        self._xkb_info = xkb_info
        self._possible_values = []

        model_values = []
        if not self._multiple_selection:
            model_values.append((None, _("Default")))

        for option_id in self._xkb_info.get_options_for_group(group_id):
            desc = self._xkb_info.description_for_option(group_id, option_id)
            model_values.append((option_id, desc))
            self._possible_values.append(option_id)

        def values_cmp_py3_wrap(f):
            ''' https://docs.python.org/3/howto/sorting.html#the-old-way-using-the-cmp-parameter '''
            class C:
                def __init__(self, obj, *args):
                    self.obj = obj
                def __lt__(self, other):
                    return f(self.obj, other.obj) < 0
                def __gt__(self, other):
                    return f(self.obj, other.obj) > 0
                def __eq__(self, other):
                    return f(self.obj, other.obj) == 0
                def __le__(self, other):
                    return f(self.obj, other.obj) <= 0
                def __ge__(self, other):
                    return f(self.obj, other.obj) >= 0
                def __ne__(self, other):
                    return f(self.obj, other.obj) != 0
            return C

        def values_cmp(xxx_todo_changeme, xxx_todo_changeme1):
            (av, ad) = xxx_todo_changeme
            (bv, bd) = xxx_todo_changeme1
            if not av:
                return -1
            elif not bv:
                return 1
            else:
                return (ad > bd) - (ad < bd)
        model_values.sort(key=values_cmp_py3_wrap(values_cmp))

        self._widgets = dict()
        for (val, name) in model_values:
            w = Gtk.CheckButton.new_with_label(name)
            if not self._multiple_selection:
                w.set_group(self._widgets.get(None))

            self._widgets[val] = w
            vbox.append(w)
            w._changed_id = w.connect('toggled', self._on_toggled)
            w._val = val

        self.widget_for_size_group = None
        self.reload()

    def reload(self):
        self._values = []
        for v in self._parent_settings.get_strv(TypingTweakGroup.XKB_GSETTINGS_NAME):
            if (v in self._possible_values):
                self._values.append(v)

        self._update_checks()

    def _update_checks(self):
        if len(self._values) > 0:
            self.set_label('<b>'+self.title+'</b>')
            self.set_use_markup(True)
        else:
            self.set_label(self.title)

        def _set_active(w, active):
            w.disconnect(w._changed_id)
            w.set_active(active)
            w._changed_id = w.connect('toggled', self._on_toggled)

        if not self._multiple_selection:
            if len(self._values) > 0:
                w = self._widgets.get(self._values[0])
                if w:
                    _set_active(w, True)
        else:
            for w in list(self._widgets.values()):
                if w._val in self._values:
                    _set_active(w, True)
                else:
                    _set_active(w, False)

    def _on_toggled(self, w):
        active = w.get_active()
        if not self._multiple_selection and active:
            for v in self._values:
                self._parent_settings.setting_remove_from_list(TypingTweakGroup.XKB_GSETTINGS_NAME, v)

        if w._val in self._values and not active:
            self._parent_settings.setting_remove_from_list(TypingTweakGroup.XKB_GSETTINGS_NAME, w._val)
        elif active and not w._val in self._values and w._val:
            self._parent_settings.setting_add_to_list(TypingTweakGroup.XKB_GSETTINGS_NAME, w._val)

class TypingTweakGroup(Gtk.Box):

    XKB_GSETTINGS_SCHEMA = "org.gnome.desktop.input-sources"
    XKB_GSETTINGS_NAME = "xkb-options"
    # grp_led is unsupported
    XKB_OPTIONS_BLACKLIST = {"grp_led", "Compose key"}

    def __init__(self):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=3)
        self._option_objects = []
        ok = False
        try:
            self._kbdsettings = GSettingsSetting(self.XKB_GSETTINGS_SCHEMA)
            self._kdb_settings_id = self._kbdsettings.connect("changed::"+self.XKB_GSETTINGS_NAME, self._on_changed)
            self._xkb_info = GnomeDesktop.XkbInfo()
            ok = True
            self.loaded = True
        except GSettingsMissingError:
            logging.info("Typing missing schema %s" % self.XKB_GSETTINGS_SCHEMA)
            self.loaded = False
        except AttributeError:
            logging.warning("Typing missing GnomeDesktop.gir with Xkb support")
            self.loaded = False
        finally:
            if ok:
                for opt in set(self._xkb_info.get_all_option_groups()) - self.XKB_OPTIONS_BLACKLIST:
                    obj = _XkbOption(opt, self._kbdsettings, self._xkb_info)
                    self._option_objects.append(obj)
                self._option_objects.sort(key=lambda item_desc: item_desc.title)
                for item in self._option_objects:
                    self.append(item)
        TweakGroup.__init__(self, _("Typing"), *self._option_objects)

        self.connect("destroy", self._on_destroy)

    def _on_changed(self, *args):
        for obj in self._option_objects:
            obj.reload()

    def _on_destroy(self, event):
        if (self._kdb_settings_id):
            self._kbdsettings.disconnect(self._kdb_settings_id)


class KeyThemeSwitcher(GSettingsSwitchTweakValue):
    def __init__(self, **options):
        GSettingsSwitchTweakValue.__init__(self,
                                           _("Emacs Input"),
                                           "org.gnome.desktop.interface",
                                           "gtk-key-theme",
                                           desc=_("Overrides shortcuts to use keybindings from the Emacs editor."),
                                           **options)

    def get_active(self):
        return "Emacs" in self.settings.get_string(self.key_name)

    def set_active(self, v):
        if v:
            self.settings.set_string(self.key_name, "Emacs")
        else:
            self.settings.set_string(self.key_name, "Default")


class OverviewShortcutTweak(Gtk.Box, _GSettingsTweak):

    def __init__(self, **options):
        name = _("Overview Shortcut")
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        _GSettingsTweak.__init__(self, name, "org.gnome.mutter", "overlay-key", loaded=_shell_loaded, **options)

        box_btn = Gtk.Box()
        box_btn.set_homogeneous(True)
        box_btn.add_css_class("linked")

        btn1 = Gtk.ToggleButton.new_with_label(_("Left Super"))
        btn2 = Gtk.ToggleButton.new_with_label(_("Right Super"))
        btn2.set_group(btn1)

        if self.settings.get_string(self.key_name) == "Super_R":
            btn2.set_active(True)
        elif self.settings.get_string(self.key_name) == "Super_L":
            btn1.set_active(True)

        btn1.connect("toggled", self.on_button_toggled, "Super_L")
        btn2.connect("toggled", self.on_button_toggled, "Super_R")

        box_btn.append(btn1)
        box_btn.append(btn2)
        build_label_beside_widget(name, box_btn, hbox=self)

    def on_button_toggled(self, button, key):
        self.settings[self.key_name] = key


class AdditionalLayoutButton(Gtk.Box, Tweak):

    def __init__(self):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=18,
                               valign=Gtk.Align.CENTER)
        Tweak.__init__(self, "extensions", "")

        btn = Gtk.Button(label=_("Additional Layout Options"), halign=Gtk.Align.END)
        btn.connect("clicked", self._on_browse_clicked)
        self.append(btn)

    def _on_browse_clicked(self, btn):
        dialog = Gtk.Dialog()
        dialog.set_title(_("Additional Layout Options"))
        dialog.set_transient_for(self.main_window)
        dialog.set_modal(True)
        dialog.set_size_request(500, 500)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_margin_top(10)
        scrolled_window.set_margin_start(10)
        box = TypingTweakGroup()
        scrolled_window.set_child(box)

        dialog.set_child(scrolled_window)
        dialog.show()


TWEAK_GROUP = TweakPreferencesPage("keyboard", _("Keyboard"),
                                      GSettingsTweakSwitchRow(_("Show Extended Input Sources"),
                          "org.gnome.desktop.input-sources",
                          "show-all-sources",
                          desc=_("Increases the choice of input sources in the Settings application."),
                          logout_required=True,),        
                                TweakPreferencesGroup(
                                _("Layout"),    "keyboard-layout", 
                            
    KeyThemeSwitcher(),
    OverviewShortcutTweak(),
    AdditionalLayoutButton(),
                                )
  
)



# Copyright (c) 2011 John Stowers
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSES/GPL-3.0



from gtweak.widgets import TweakPreferencesPage, GSettingsTweakSpinRow, GSettingsTweakFontRow, TweakPreferencesGroup, TweaksCheckGroupActionRow

class FontHintingTweak(TweaksCheckGroupActionRow):

    def __init__(self, **options):
        TweaksCheckGroupActionRow.__init__(self,
                             title=_("Hinting"),
                             setting="org.gnome.desktop.interface",
                             key_name="font-hinting"
        )

        self.add_row(title=_("Full"), key_name="full")
        self.add_row(title=_("Medium"), key_name="medium")
        self.add_row(title=_("Slight"), key_name="slight")
        self.add_row(title=_("None"), key_name="none")

class FontAliasingTweak(TweaksCheckGroupActionRow):

    def __init__(self, **options):
        TweaksCheckGroupActionRow.__init__(self,
                             title=_("Antialiasing"),
                             setting="org.gnome.desktop.interface",
                             key_name="font-antialiasing"
        )
      
        self.add_row(title=_("Subpixel (for LCD screens)"), key_name="rgba")
        self.add_row(title=_("Standard (grayscale)"), key_name="grayscale")
        self.add_row(title=_("None"), key_name="none")


TWEAK_GROUP = TweakPreferencesPage("fonts", _("Fonts"),
                                TweakPreferencesGroup(
                                    _("Preferred Fonts"), "preferred-fonts",
    GSettingsTweakFontRow(_("Interface Text"),"org.gnome.desktop.interface", "font-name"),
    GSettingsTweakFontRow(_("Document Text"), "org.gnome.desktop.interface", "document-font-name"),
    GSettingsTweakFontRow(_("Monospace Text"), "org.gnome.desktop.interface", "monospace-font-name"),
                                ),
                                 TweakPreferencesGroup(
                                     _("Rendering"), "font-rendering",
    FontHintingTweak(),
    FontAliasingTweak(),
                                 ),
                                  TweakPreferencesGroup( _("Size"), "font-size",
    GSettingsTweakSpinRow(_("Scaling Factor"),
      "org.gnome.desktop.interface", "text-scaling-factor",
      adjustment_step=0.01, digits=2),
                                  ),
)




# Copyright (c) 2011 John Stowers
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSES/GPL-3.0

import logging
import os.path
import subprocess
from typing import Set, Optional

from gi.repository import Adw, Gtk, Gdk, Gio, GObject, GLib

from gtweak.tweakmodel import Tweak, TweakGroup
from gtweak.utils import AutostartManager, AutostartFile


def _image_from_gicon(gicon):
    image = Gtk.Image.new_from_gicon(gicon)
    image.set_icon_size(Gtk.IconSize.LARGE)
    return image


class _AppChooserRow(Gtk.ListBoxRow):
    """The row in the appchooser to show desktop files """

    def __init__(self, app_info: Gio.AppInfo, is_running: bool, **kwargs):
        super().__init__(**kwargs)
        self.app_name = app_info.get_name().lower()
        self.is_running = is_running
        self.app_info = app_info

        vbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        vbox.set_margin_top(10)
        vbox.set_margin_bottom(10)
        vbox.set_margin_start(10)
        vbox.set_margin_end(10)

        icon = app_info.get_icon()
        if icon:
            img = _image_from_gicon(icon)
            img.set_hexpand(False)
        else:
            img = Gtk.Image.new_from_icon_name("application-x-executable")
            img.props.icon_size = Gtk.IconSize.LARGE
        app_name = app_info.get_name()
        lbl = Gtk.Label(label=app_name, hexpand=True,
                        halign=Gtk.Align.START, wrap=True)
        vbox.append(img)
        vbox.append(lbl)

        if is_running:
            lbl_running = Gtk.Label(label=_("running"))
            vbox.append(lbl_running)
        self.set_child(vbox)


class _AppChooser(Gtk.Dialog):
    """Presents a dialog to select a desktop file """

    def __init__(self, main_window, running_exes, startup_apps):
        uhb = Gtk.Settings.get_default().props.gtk_dialogs_use_header
        Gtk.Dialog.__init__(self, title=_("Select Application"), use_header_bar=uhb)

        self._running = {}
        self._all = {}

        # Build header bar buttons
        self.add_button(_("_Close"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Add"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)

        self.entry = Gtk.SearchEntry(
            placeholder_text=_("Search Applications…"))
        self.entry.set_width_chars(30)
        self.entry.props.activates_default = True
        self.entry.connect("search-changed", self._on_search_entry_changed)

        self.searchbar = Gtk.SearchBar()
        self.searchbar.set_child(self.entry)
        self.searchbar.set_hexpand(True)
        self.searchbar.set_key_capture_widget(self)

        lb = Gtk.ListBox()
        lb.set_activate_on_single_click(False)
        lb.set_sort_func(self._list_sort_func, None)
        lb.set_filter_func(self._list_filter_func, self.entry)
        lb.connect("row-activated",
                   lambda b, r: self.response(Gtk.ResponseType.OK) if r.get_mapped() else None)
        lb.connect("row-selected", self._on_row_selected)

        apps = filter(lambda _app: _app.should_show() and _app.get_id() not in startup_apps,
                      Gio.app_info_get_all())

        for app in apps:
            running = app.get_executable() in running_exes
            if not app.get_name():
                continue
            w = _AppChooserRow(app, running)
            if w:
                lb.append(w)

        sw = Gtk.ScrolledWindow(vexpand=True, margin_top=2, margin_bottom=2,
                                margin_start=2, margin_end=2)
        sw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        sw.set_child(lb)

        header_bar = self.get_header_bar()
        if header_bar:
            searchbtn = Gtk.ToggleButton(valign=Gtk.Align.CENTER)
            searchbtn.set_icon_name("edit-find-symbolic")
            header_bar.pack_end(searchbtn)
            self._binding = searchbtn.bind_property(
                "active", self.searchbar,
                "search-mode-enabled", GObject.BindingFlags.BIDIRECTIONAL)

        self.get_content_area().append(self.searchbar)
        self.get_content_area().append(sw)
        self.set_modal(True)
        self.set_transient_for(main_window)
        self.set_size_request(320, 300)

        self.listbox = lb
        self._setup_shortcut()

    def _setup_shortcut(self):
        s_trigger = Gtk.ShortcutTrigger.parse_string("<primary>f")
        s_action = Gtk.CallbackAction.new(lambda w, a, s: s.set_search_mode(True), self.searchbar)

        if s_trigger and s_action:
            shortcut = Gtk.Shortcut(trigger=s_trigger, action=s_action)
            self.add_shortcut(shortcut)

    @staticmethod
    def _list_sort_func(row_a: _AppChooserRow, row_b: _AppChooserRow, _):
        a_isrunning = row_a.is_running
        b_isrunning = row_b.is_running

        if a_isrunning and not b_isrunning:
            return -1
        elif not a_isrunning and b_isrunning:
            return 1
        else:
            aname = row_a.app_name
            bname = row_b.app_name

            if aname < bname:
                return -1
            elif aname > bname:
                return 1
            else:
                return 0

    @staticmethod
    def _list_filter_func(row: _AppChooserRow, entry: Gtk.SearchEntry) -> bool:
        txt = entry.get_text().lower()
        if txt in row.app_name:
            return True
        return False

    def _on_search_entry_changed(self, editable):
        self.listbox.invalidate_filter()
        selected = self.listbox.get_selected_row()
        if selected and selected.get_mapped():
            self.set_response_sensitive(Gtk.ResponseType.OK, True)
        else:
            self.set_response_sensitive(Gtk.ResponseType.OK, False)

    def _on_row_selected(self, box, row):
        if row and row.get_mapped():
            self.set_response_sensitive(Gtk.ResponseType.OK, True)
        else:
            self.set_response_sensitive(Gtk.ResponseType.OK, False)

    def get_selected_appinfo(self) -> Optional[Gio.AppInfo]:
        row: _AppChooserRow = self.listbox.get_selected_row()

        if row is not None:
            return row.app_info
        return None


class _StartupAppRowTweak(Adw.ActionRow, Tweak):

    def __init__(self, desktop_info: Gio.AppInfo, **options):
        Adw.PreferencesRow.__init__(self)
        Tweak.__init__(self, desktop_info.get_name(), desktop_info.get_description(), **options)

        icon = desktop_info.get_icon()
        if icon:
            app_icon = _image_from_gicon(icon)
        else:
            app_icon = Gtk.Image.new_from_icon_name("image-missing")
            app_icon.set_icon_size(Gtk.IconSize.LARGE)

        self.btn = Gtk.Button(icon_name="edit-delete-symbolic")
        self.btn.set_tooltip_text(_("Remove"))

        self.btn.add_css_class("flat")
        self.btn.set_vexpand(False)
        self.btn.set_valign(Gtk.Align.CENTER)

        self.set_title(desktop_info.get_name())
        self.add_prefix(app_icon)
        self.add_suffix(self.btn)
        self.app_info = desktop_info

        controller_key = Gtk.EventControllerKey()
        self.add_controller(controller_key)
        controller_key.connect("key-pressed", self._on_key_press_event)

    def _on_key_press_event(self, _, keyval: int, _1, _2):
        if keyval in [Gdk.KEY_Delete, Gdk.KEY_KP_Delete, Gdk.KEY_BackSpace]:
            self.btn.activate()
            return True
        return False


class AutostartTweakGroup(Adw.PreferencesPage, TweakGroup):

    def __init__(self, *tweaks, **options):
        name: str = _("Startup Applications")
        desc: str = _("Startup applications are automatically started when you log in.")
        Adw.PreferencesPage.__init__(self)
        TweakGroup.__init__(self, "startup-applications", name, **options)

        self.tweaks = [Tweak(name, desc, **options)]

        pregroup = Adw.PreferencesGroup()
        # Preferencee Group Header
        pregroup.set_title(name)
        pregroup.set_description(desc)

        self.btn_add_startup = Gtk.Button(valign=Gtk.Align.CENTER)
        self.btn_add_startup.set_icon_name("list-add-symbolic")
        self.btn_add_startup.add_css_class("flat")
        pregroup.set_header_suffix(self.btn_add_startup)

        # Body
        self.stack = Gtk.Stack()
        self.status_page = Adw.StatusPage()
        self.pg_startup_apps = Adw.PreferencesGroup()
        self._setup_stack_view()

        self._startup_dapps = self._get_startup_desktop_files()
        self._setup_startup_app_row()

        self.__init_connections()
        pregroup.add(self.stack)
        self.add(pregroup)
        self._set_visible_page()

    def _setup_stack_view(self):
        self.stack.set_vexpand(True)
        self.stack.set_hhomogeneous(True)
        self.stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)

        # Empty Page
        self.status_page.set_icon_name("application-x-executable-symbolic")
        self.status_page.set_title(_("No Startup Applications"))
        self.status_page.set_description(_("Add a startup application"))
        self.status_page.add_css_class("dim-label")

        self.stack.add_child(self.status_page)
        self.stack.add_child(self.pg_startup_apps)

    def _setup_startup_app_row(self):
        """ Add a row for each autostart applications existing"""

        dfiles = self._startup_dapps
        for dfile in dfiles:
            app_row = _StartupAppRowTweak(dfile)
            app_row.btn.connect("clicked", self._on_remove_clicked, app_row)
            self.pg_startup_apps.add(app_row)
            self.tweaks.append(app_row)

    def _set_visible_page(self):
        if len(self._startup_dapps) > 0:
            self.stack.set_visible_child(self.pg_startup_apps)
        else:
            self.stack.set_visible_child(self.status_page)

    def _on_add_clicked(self, _: Gtk.Button):
        def _on_response_appchooser(chooser: _AppChooser, response_id: int):
            if response_id == Gtk.ResponseType.OK:
                appinfo = chooser.get_selected_appinfo()

                if appinfo:
                    AutostartFile(appinfo).update_start_at_login(True)
                    arow_app_row = _StartupAppRowTweak(appinfo)
                    arow_app_row.btn.connect("clicked", self._on_remove_clicked, arow_app_row)

                    self.pg_startup_apps.add(arow_app_row)
                    self._startup_dapps.add(appinfo)
                    self._set_visible_page()
            chooser.destroy()

        startup_app_ids = tuple(map(lambda x: x.get_id(), self._startup_dapps))

        Gio.Application.get_default().mark_busy()
        a = _AppChooser(self.main_window, self._get_running_executables(), startup_app_ids)
        a.connect("response", _on_response_appchooser)
        Gio.Application.get_default().unmark_busy()
        a.present()

    def _on_remove_clicked(self, btn, app_row: _StartupAppRowTweak):
        app_info = app_row.app_info
        AutostartFile(app_info).update_start_at_login(False)

        self.pg_startup_apps.remove(app_row)
        self._startup_dapps.remove(app_info)
        self._set_visible_page()

    def __init_connections(self):
        self.btn_add_startup.connect("clicked", self._on_add_clicked)

    @staticmethod
    def _get_startup_desktop_files():
        asm = AutostartManager()
        autostart_files = asm.get_user_autostart_files()
        dfiles = set()

        for file in autostart_files:
            try:
                dappinfo = Gio.DesktopAppInfo.new_from_filename(file)
            except TypeError:
                logging.warning(f"Error loading desktop file: {file}")
            else:
                if not AutostartFile(dappinfo).is_start_at_login_enabled():
                    continue
                dfiles.add(dappinfo)

        return dfiles

    @staticmethod
    def _get_running_executables() -> Set[str]:
        exes = set()
        cmd = subprocess.Popen([
            'ps', '-e', '-w', '-w', '-U',
            str(os.getuid()), '-o', 'cmd'],
            stdout=subprocess.PIPE)
        out = cmd.communicate()[0]
        for process in out.decode('utf8').split('\n'):
            exe = process.split(' ')[0]
            if exe and exe[0] != '[':  # kernel process
                exes.add(os.path.basename(exe))

        return exes

TWEAK_GROUP = AutostartTweakGroup()



# Copyright (c) 2024 Evan Welsh
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSES/GPL-3.0

from gi.repository import GUdev

# Inspired by panels/common/gsd-device-manager.c in GNOME Settings
# https://gitlab.gnome.org/GNOME/gnome-control-center/-/blob/6d2add0e30538692c151e5fa0bd94ae9bece7690/panels/common/gsd-device-manager.c

UDEV_ID_INPUT_MOUSE = "ID_INPUT_MOUSE"
UDEV_ID_INPUT_KEYBOARD = "ID_INPUT_KEYBOARD"
UDEV_ID_INPUT_TOUCHPAD = "ID_INPUT_TOUCHPAD"
UDEV_ID_INPUT_POINTING_STICK = "ID_INPUT_POINTINGSTICK"

UDEV_IDS = [
    UDEV_ID_INPUT_MOUSE,
    UDEV_ID_INPUT_KEYBOARD,
    UDEV_ID_INPUT_TOUCHPAD,
    UDEV_ID_INPUT_POINTING_STICK,
]

def udev_device_is_evdev(device):
    device_file = device.get_device_file()
    if device_file is None or "/event" not in device_file:
        return False
    
    return device.get_property_as_boolean("ID_INPUT")


def get_input_devices():
    udev_client = GUdev.Client()
    devices = udev_client.query_by_subsystem ("input")
    
    return [device for device in devices if udev_device_is_evdev(device)]


def udev_device_get_device_types (device):
    types = set()

    for id in UDEV_IDS:
        if device.get_property_as_boolean(id):
            types.add(id)

    return types


def udev_device_id_is_present(id):
    for device in get_input_devices():
        if id in udev_device_get_device_types(device):
              return True

    return False


def pointing_stick_is_present():
    return udev_device_id_is_present(UDEV_ID_INPUT_POINTING_STICK)


def touchpad_is_present():
    return udev_device_id_is_present(UDEV_ID_INPUT_TOUCHPAD)


# Copyright (c) 2011 John Stowers
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSES/GPL-3.0

from gi.repository import Adw
from gi.repository import Gtk
from gi.repository import Gio
from gi.repository import GLib

import gtweak
from gtweak.defs import VERSION
from gtweak.tweakmodel import TweakModel
from gtweak.tweakview import Window
from gtweak.utils import SchemaList
from gtweak.gshellwrapper import GnomeShellFactory

class ExtensionNotice(Gtk.MessageDialog):
    def __init__(self, modal, transient_for):
        Gtk.Dialog.__init__(self, modal=modal, transient_for=transient_for)

        self.add_button(_("_Continue"), Gtk.ResponseType.NONE)

        self.set_markup("<b>{0}</b>".format(_("Extensions Has Moved")))

        self.props.secondary_use_markup = True
        self.props.secondary_text = "{0}\n\n{1}".format(
            # Translators: Placeholder will be replaced with "GNOME Extensions" in active link form
            _("Extensions management has been moved to {0}.").format(
                '<a href="https://gitlab.gnome.org/GNOME/gnome-shell/-/blob/HEAD/subprojects/extensions-app/README.md">GNOME Extensions</a>',
            ),
            # Translators: Placeholder will be replaced with "Flathub" in active link form
            _("We recommend downloading GNOME Extensions from {0} if your distribution does not include it.").format(
                '<a href="https://flathub.org/apps/details/org.gnome.Extensions">Flathub</a>'
            )
        )

_application = None

def get_application():
    return _application

def get_window():
    return _application.win

class GnomeTweaks(Adw.Application):

    def __init__(self):
        global _application

        _application = self

        GLib.set_application_name(_("GNOME Tweaks"))
        super().__init__(application_id=gtweak.APP_ID)
        self.win = None

        self._settings = Gio.Settings.new('org.gnome.tweaks')

    def do_activate(self):
        if not self.win:
            model = TweakModel()
            self.win = Window(self, model)
        self.win.present()

        if self._settings.get_boolean('show-extensions-notice'):
            self.show_extensions_notice()
            self._settings.set_boolean('show-extensions-notice', False)

    def do_startup(self):
        Adw.Application.do_startup(self)

        self.set_accels_for_action("window.close", ["<primary>w"])
        self._create_action("quit", self.quit_app, ["<primary>q"])
        self._create_action("about", self.about_cb)
        self._create_action("reset", self.reset_cb)

    def quit_app(self, *_):
        self.quit()

    def reset_cb(self, action, parameter):
        def _on_dialog_response(_dialog, response_type):
            if response_type == Gtk.ResponseType.OK:
                SchemaList.reset()

            _dialog.destroy()

        dialog = Gtk.MessageDialog(transient_for=self.win,
                                   modal=True,
                                   message_type=Gtk.MessageType.QUESTION,
                                   buttons=Gtk.ButtonsType.OK_CANCEL,
                                   text=_("Reset to Defaults"),
                                   secondary_text=_("Reset all tweaks settings to the original default state?"))

        dialog.connect("response", _on_dialog_response)
        dialog.present()

    def about_cb(self, action, parameter):
        _shell = GnomeShellFactory().get_shell()
        if _shell is not None:
            if _shell.mode == "user":
                about_comment = f'{_("GNOME Shell")} {_shell.version}'

            else:
                about_comment = (_("GNOME Shell") + " %s " + _("(%s mode)")) % \
                    (_shell.version, _shell.mode)
        else:
            about_comment = _("GNOME Shell is not running")

        about_comment += f'\n{_("GTK")} {Gtk.get_major_version()}.{Gtk.get_minor_version()}.{Gtk.get_micro_version()}'

        AUTHORS = [
                "John Stowers <john.stowers@gmail.com>"
                ]

        aboutdialog = Adw.AboutWindow(
            application_name=GLib.get_application_name(),
            application_icon=gtweak.APP_ID,
            comments=about_comment,
            copyright="Copyright © 2011 - 2013 John Stowers.",
            developer_name="John Stowers",
            # TRANSLATORS: Add your name/nickname here (one name per line),
            # they will be displayed in the "about" dialog
            translator_credits=_("translator-credits"),
            developers=AUTHORS,
            transient_for=self.win,
            version=VERSION,
            website="https://wiki.gnome.org/Apps/Tweaks",
            issue_url="https://gitlab.gnome.org/GNOME/gnome-tweaks/-/issues",
            license_type=Gtk.License.GPL_3_0
        )

        aboutdialog.present()

    def show_extensions_notice(self):
        extensionsdialog = ExtensionNotice(
            modal=True,
            transient_for=self.win
        )

        extensionsdialog.connect("response", lambda _dialog, _: _dialog.destroy())
        extensionsdialog.show()

    def _create_action(self, name, callback, shortcuts=None):
        """Add an application action.

        :param name: the name of the action
        :param callback: the function to be called when the action is activated
        :param shortcuts:
        :return: an optional list of accelerators
        """
        action = Gio.SimpleAction.new(name, None)
        action.connect("activate", callback)
        self.add_action(action)
        if shortcuts:
            self.set_accels_for_action(f"app.{name}", shortcuts)



# Copyright (c) 2011 John Stowers
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSES/GPL-3.0

import os.path
from typing import List

from gi.repository import Adw, Gtk, Gdk, GObject
import gtweak
import gtweak.widgets 
import gtweak.tweakmodel
from gtweak.tweakmodel import TweakModel, string_for_search

from gtweak.tweaks.tweak_group_appearance import TWEAK_GROUP as AppearanceTweaks
from gtweak.tweaks.tweak_group_font import TWEAK_GROUP as FontTweaks
from gtweak.tweaks.tweak_group_mouse import TWEAK_GROUP as MouseTweaks
from gtweak.tweaks.tweak_group_keyboard import TWEAK_GROUP as KeyboardTweaks
from gtweak.tweaks.tweak_group_sound import TWEAK_GROUP as SoundTweaks, show_sound_tweaks
from gtweak.tweaks.tweak_group_windows import TWEAK_GROUP as WindowTweaks
from gtweak.tweaks.tweak_group_startup import TWEAK_GROUP as StartupApplicationTweaks

tweaks = [ 
    MouseTweaks,
    KeyboardTweaks,
    FontTweaks,
    AppearanceTweaks,
    WindowTweaks,
    StartupApplicationTweaks,
    SoundTweaks
]


@Gtk.Template(filename=os.path.join(gtweak.PKG_DATA_DIR, "tweaks.ui"))
class Window(Adw.ApplicationWindow):
    __gtype_name__ = "GTweakWindow"

    main_box = Gtk.Template.Child()
    listbox = Gtk.Template.Child()
    right_box = Gtk.Template.Child()
    header = Gtk.Template.Child()
    main_content_scroll = Gtk.Template.Child()
    main_leaflet = Gtk.Template.Child()
    main_stack = Gtk.Template.Child()
    left_box = Gtk.Template.Child()
    searchbar = Gtk.Template.Child()
    entry = Gtk.Template.Child()
    list_box_row_sound = Gtk.Template.Child()

    def __init__(self, app: Adw.Application, model: TweakModel):
        super().__init__(application=app, show_menubar=False)
        self.set_default_size(980, 640)
        self.set_size_request(-1, 300)
        self.set_icon_name(gtweak.APP_ID)

        for tweak in tweaks:
            model.add_tweak_group(tweak, self)
            if tweak.name:
              self.main_stack.add_named(tweak, tweak.name)

        self._setup_header()
        self._setup_sidebar()
        self._setup_mainbox()
        self._setup_sizegroups()

        self._load_css()
        self._model = model

        self._setup_shortcut()
        self.searchbar.set_key_capture_widget(self)


    def _setup_header(self):
        self.hsize_group = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)
        self.menu_btn = Gtk.MenuButton()

        header = self.header
        header.set_transition_type(Adw.LeafletTransitionType.SLIDE)

        left_header = Adw.HeaderBar()
        self._left_header = left_header
        left_header.set_show_end_title_buttons(False)
        left_header.add_css_class("titlebar")
        left_header.add_css_class("tweak-titlebar-left")

        self.search_btn = Gtk.ToggleButton()  # Search Button
        self.search_btn.set_icon_name("edit-find-symbolic")
        self.search_btn.bind_property("active", self.searchbar, "search-mode-enabled",
                                      GObject.BindingFlags.BIDIRECTIONAL)
        self.search_btn.set_valign(Gtk.Align.CENTER)
        self.search_btn.add_css_class("image-button")
        left_header.pack_start(self.search_btn)

        lbl = Adw.WindowTitle.new(_("Tweaks"), "")  # Left label
        left_header.set_title_widget(lbl)

        self.builder = Gtk.Builder()
        assert(os.path.exists(gtweak.PKG_DATA_DIR))
        filename = os.path.join(gtweak.PKG_DATA_DIR, 'shell.ui')
        self.builder.add_from_file(filename)

        appmenu = self.builder.get_object('appmenu')  # Left AppMenu
        self.menu_btn.set_icon_name("open-menu-symbolic")
        self.menu_btn.set_menu_model(appmenu)
        left_header.pack_end(self.menu_btn)

        # Right Header
        right_header = Adw.HeaderBar(hexpand=True)
        self._right_header = right_header
        right_header.set_show_start_title_buttons(False)

        right_header.add_css_class("titlebar")
        right_header.add_css_class("tweak-titlebar-right")

        self._group_titlebar_widget = None

        self.title = Adw.WindowTitle.new("", "")
        right_header.set_title_widget(self.title)

        self.back_button = Gtk.Button.new_from_icon_name("go-previous-symbolic")
        self.back_button.connect("clicked", self._on_back_clicked)
        self.back_button.props.visible = header.props.folded
        right_header.pack_start(self.back_button)

        header.bind_property("folded", self.back_button, "visible",
                                        GObject.BindingFlags.SYNC_CREATE)
        header.bind_property("folded", right_header, "show-start-title-buttons")
        header.bind_property("folded", left_header, "show-end-title-buttons")

        lbl = Adw.WindowTitle.new(_("Tweaks"), "")
        left_header.set_title_widget(lbl)

        self.builder = Gtk.Builder()
        assert os.path.exists(gtweak.PKG_DATA_DIR)
        filename = os.path.join(gtweak.PKG_DATA_DIR, "shell.ui")
        self.builder.add_from_file(filename)

        header.append(left_header)
        header.get_page(left_header).set_name("sidebar")
        header.append(Gtk.Separator(orientation=Gtk.Orientation.VERTICAL))
        header.append(right_header)
        header.get_page(right_header).set_name("content")

        self.hsize_group.add_widget(left_header)

        return header

    def _setup_mainbox(self):
        # self.main_box.set_size_request(540, -1)

        # self.main_stack.add_css_class("background")
        self.main_stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)

        self.main_leaflet.bind_property("visible-child-name", self.header, "visible-child-name", GObject.BindingFlags.SYNC_CREATE)

    def _setup_sidebar(self):
        self.entry.placeholder_text=_("Search Tweaks…")
        if Gtk.check_version(3, 22, 20) is None:
            self.entry.set_input_hints(Gtk.InputHints.NO_EMOJI)
        self.entry.connect("search-changed", self._on_search)

        self.listbox.connect("row-selected", self._on_select_row)

        if not show_sound_tweaks:
            self.listbox.remove(self.list_box_row_sound)

    def _setup_sizegroups(self):
        start_pane_size_group = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)
        start_pane_size_group.add_widget(self.left_box)
        start_pane_size_group.add_widget(self._left_header)

        end_pane_size_group = Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)
        end_pane_size_group.add_widget(self.right_box)
        end_pane_size_group.add_widget(self._right_header)

    def _setup_shortcut(self):
        s_trigger = Gtk.ShortcutTrigger.parse_string("<primary>f")
        s_action = Gtk.CallbackAction.new(lambda w, a, s: s.set_search_mode(True), self.searchbar)

        if s_trigger and s_action:
            shortcut = Gtk.Shortcut(trigger=s_trigger, action=s_action)
            self.add_shortcut(shortcut)

    def _load_css(self):
        if gtweak.defs.IS_DEVEL:
            self.add_css_class('devel')
        css_provider = Gtk.CssProvider()
        css_provider.load_from_path(
            os.path.join(gtweak.PKG_DATA_DIR, 'shell.css'))
        display = Gdk.Display.get_default()
        self.get_style_context().add_provider_for_display(display, css_provider,
                                            Gtk.STYLE_PROVIDER_PRIORITY_USER)

    @staticmethod
    def _list_filter_func(row, user_data: List[str]):
        name = row.props.tweakname
        if name in user_data:
            return row

    def _after_key_press(self, widget, event):
        if not self.search_btn.get_active() or not self.entry.is_focus():
            if self.entry.im_context_filter_keypress(event):
                self.search_btn.set_active(True)
                self.entry.grab_focus()

                # Text in entry is selected, deselect it
                l = self.entry.get_text_length()
                self.entry.select_region(l, l)
                return True

        return False

    def _on_list_changed(self, group):
        self.listbox.set_filter_func(self._list_filter_func, group)
        selected = self.listbox.get_selected_row()
        if not selected:
            return
        selected_tweakname = selected.props.tweakname
        if group and selected_tweakname not in group:
            index = sorted(self._model._tweak_group_names.keys()).index(group[0])
            row = self.listbox.get_row_at_index(index)
            self.listbox.select_row(row)
            self.entry.grab_focus()

    def _on_search(self, entry: Gtk.SearchEntry):
        txt = string_for_search(entry.get_text())
        group = self._model.search_matches(txt)
        self._on_list_changed(group)

    def _on_select_row(self, _, row: Gtk.ListBoxRow):
        if row:
            group = row.props.tweakname

            for tweak in tweaks:
              if tweak.name == group:

                self.main_stack.set_visible_child_name(tweak.name)
                self.title.set_title(tweak.title)
                self.main_leaflet.set_visible_child_name("content")

    def _on_find_toggled(self, _):
        self.searchbar.set_search_mode(not self.searchbar.get_search_mode())

    def _on_back_clicked(self, *_):
        # Clear the page selection when going back to allow
        # re-selecting it.
        self.listbox.unselect_all()

        self.main_leaflet.set_visible_child_name("sidebar")

    def show_only_tweaks(self, tweaks):
        for t in self._model.tweaks:
            if not isinstance(t, Gtk.Widget):
                continue
            if t in tweaks:
                t.show()
            else:
                t.hide()

